#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ------------------------------------------------
# COPYRIGHT (C) 2014-2024 Mitsuo KONDOU.
# This software is released under the not MIT License.
# Permission from the right holder is required for use.
# https://github.com/ClipnGPT
# Thank you for keeping the rules.
# ------------------------------------------------

import json

import os
import sys
import time
import pygame

import subprocess

from gtts import gTTS

import requests
import wave

import platform    # platform.system().lower() #windows,darwin,linux
if (platform.system().lower() == 'windows'):
    import pythoncom
    import win32com.client

qPath_work  = 'temp/_work/'



class tts_class:
    def __init__(self, ):
        self.mixer_enable    = False
        self.voicevox_enable = False
        self.voicevox_host   = 'localhost'
        self.voicevox_port   = 50021

        # VOICEVOX チェック
        try:
            host = self.voicevox_host
            port = self.voicevox_port
            res = requests.get(
                    f'http://{host}:{port}/version',
                    timeout=(2, 5)
            )
            if (res.status_code == 200):
                self.voicevox_enable  = True
                self.voicevox_version = res.json()
                print('Voicevox ready! (' + self.voicevox_version + ')')
        except:
            pass

    def google_tts(self, outText='おはよう', outLang='ja', outSpeaker='google', outGender='female', outFile='temp/_work/tts.mp3', ):
        outFile = outFile[:-4] + '.mp3'

        # 引数チェック
        text = outText.strip()
        if (text == '') or (text == '!'):
            return False, None

        lang   = outLang
        gender = outGender.lower()

        # 音声合成
        try:
            tts = gTTS(text=text, lang=lang, slow=False)
            tts.save(outFile)
            time.sleep(0.10)
        except Exception as e:
            print(e)
            return False, None

        # 結果確認
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 4096):
                os.remove(outFile)
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 44):
                os.remove(outFile)
            else:
                return outFile, 'google'

        return False, None

    def winos_tts(self, outText='おはよう', outLang='ja', outSpeaker='winos', outGender='female', outFile='temp/_work/tts.wav'):
        outFile = outFile[:-4] + '.wav'

        if (platform.system().lower() != 'windows'):
            return False, None

        # 引数チェック
        text = outText.strip()
        if (text == '') or (text == '!'):
            return False, None

        #ja-JP:日本語〇,    en-US:英語〇,
        #ar-AR:アラビア語x, es-ES:スペイン語×, de-DE:ドイツ語×
        #fr-FR:フランス語〇,it-IT:イタリア語×, pt-BR:ポルトガル語×
        #ru-RU:ロシア語×,  tr-TR:トルコ語×,   uk-UK:ウクライナ語×
        #zh-CN:中国語×,    kr-KR:韓国語×

        gender = outGender.lower()

        lang = ''
        if   (outLang == 'ja') or (outLang == 'ja-JP'):
            lang = 'ja-JP'
        elif (outLang == 'en') or (outLang == 'en-US'):
            lang = 'en-US'
        elif (outLang == 'ar'):
            lang = 'ar-AR'
        elif (outLang == 'es'):
            lang = 'es-ES'
        elif (outLang == 'de'):
            lang = 'de-DE'
        elif (outLang == 'fr'):
            lang = 'fr-FR'
        elif (outLang == 'it'):
            lang = 'it-IT'
        elif (outLang == 'pt'):
            lang = 'pt-BR'
        elif (outLang == 'ru'):
            lang = 'ru-RU'
        elif (outLang == 'tr'):
            lang = 'tr-TR'
        elif (outLang == 'uk'):
            lang = 'uk-UK'
        elif (outLang == 'zh') or (outLang == 'zh-CN'):
            lang = 'zh-CN'
        elif (outLang == 'kr'):
            lang = 'kr-KR'
        else:
            lang = outLang

        # 音声合成
        try:

            # MS Windows
            stml  = '<speak version="1.0" xmlns="http://www.w3.org/2001/10/synthesis" xml:lang="en-US">'
            if (lang == 'ja-JP'):
                stml += '<prosody rate="1.0">'
            stml += '<voice xml:lang="' + lang + '" gender="' + gender + '">'
            stml += text
            stml += '</voice>'
            if (lang == 'ja-JP'):
                stml += '</prosody>'
            stml += '</speak>'

            pythoncom.CoInitialize()

            try:
                engine = win32com.client.Dispatch('SAPI.SpVoice')
                #engine.Speak(stml)

            except Exception as e:
                print(lang, gender, )
                print(e)
                return False, None

            try:
                stream = win32com.client.Dispatch('SAPI.SpFileStream')
                stream.open(outFile, 3, False)
                #for speaker in engine.GetAudioOutputs():
                #    print(speaker.GetDescription())

                engine.AudioOutputStream = stream
                engine.Speak(stml)

                stream.close()
                time.sleep(0.10)
            except Exception as e:
                print(lang, gender, )
                print(e)
                return False, None

            engine = None
            stream = None

            pythoncom.CoUninitialize()

        except Exception as e:
            print(e)
            return False, None

        # 結果確認
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 4096):
                os.remove(outFile)
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 44):
                os.remove(outFile)
            else:
                return outFile, 'winos'

        return False, None

    def macos_tts(self, outText='おはよう', outLang='ja', outSpeaker='winos', outGender='female', outFile='temp/_work/tts.wav'):
        wrkFile = outFile[:-4] + '.wave'
        outFile = outFile[:-4] + '.wav'

        if (os.path.isfile(wrkFile)):
            try:
                os.remove(wrkFile)
            except Exception as e:
                print(e)

        if (platform.system().lower() != 'darwin'):
            return False, None

        # 引数チェック
        text = outText.strip()
        if (text == '') or (text == '!'):
            return False, None

        #ja-JP:日本語,    en-US:英語,
        #ar-SA:アラビア語, es-ES:スペイン語, de-DE:ドイツ語
        #fr-FR:フランス語, it-IT:イタリア語, pt-PT:ポルトガル語
        #ru-RU:ロシア語,  tr-TR:トルコ語,   uk-UK:ウクライナ語×
        #zh-CN:中国語,    ko-KR:韓国語

        gender = outGender.lower()

        voice = ''
        rate  = ''
        if   (outSpeaker.lower() == 'kyoko'):
                voice = 'Kyoko'
                rate  = '200'
        elif (outSpeaker.lower() == 'otoya'):
                voice = 'Otoya'
        elif (outLang == 'ja') or (outLang == 'ja-JP'):
            if (gender != 'male'):
                voice = 'Kyoko'
                #voice = 'O-Ren'
                #voice = ''
                rate  = '200'
            else:
                voice = 'Otoya'
                #voice = 'Hattori'
        elif (outLang == 'en') or (outLang == 'en-US'):
            if (gender != 'male'):
                voice = 'Ava'
            else:
                voice = 'Tom'
        elif (outLang == 'ar'):
            voice = 'Laila'
        elif (outLang == 'es'):
            voice = 'Monica'
        elif (outLang == 'de'):
            voice = 'Anna'
        elif (outLang == 'fr'):
            voice = 'Aurelie'
        elif (outLang == 'it'):
            voice = 'Federica'
        elif (outLang == 'pt'):
            voice = 'Joana'
        elif (outLang == 'ru'):
            voice = 'Milena'
        elif (outLang == 'tr'):
            voice = 'Yelda'
        elif (outLang == 'uk'):
            voice = ''
        elif (outLang == 'zh') or (outLang == 'zh-CN'):
            voice = 'Ting-Ting'
        elif (outLang == 'ko'):
            voice = 'Yuna'

        # 音声合成
        try:
            if (voice != ''):
                if (rate != ''):
                    say = subprocess.Popen(['say', '-v', voice, '"' + outText + '"', '-r', str(rate), '-o', wrkFile, ], )
                else:
                    say = subprocess.Popen(['say', '-v', voice, '"' + outText + '"', '-o', wrkFile, ], )
            else:
                if (rate != ''):
                    say = subprocess.Popen(['say', '"' + outText + '"', '-r', str(rate), '-o', wrkFile, ], )
                else:
                    say = subprocess.Popen(['say', '"' + outText + '"', '-o', wrkFile, ], )

            # 最大10s
            say.wait(timeout=10)
            say.terminate()
            say = None

        except Exception as e:
            print(e)
            return False, None

        # 結果確認
        if (os.path.isfile(wrkFile)):
            os.rename(wrkFile, outFile)
            
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 4096):
                os.remove(outFile)
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 44):
                os.remove(outFile)
            else:
                return outFile, 'macos'

        return False, None

    def voicevox_tts(self, outText='おはよう', outLang='ja', outSpeaker='voicevox', outGender='female', outFile='temp/_work/tts.wav', ):
        outFile = outFile[:-4] + '.wav'

        if (self.voicevox_enable != True):
            return False, None

        # 引数チェック
        text = outText.strip()
        if (text == '') or (text == '!'):
            return False, None

        speaker_id = 20 #もち子
        if (outGender.lower() == 'male') or (outGender.find('男') >= 0):
            speaker_id = 21 #龍星

        if   (outSpeaker.lower() == 'voicevox'):
            pass
        elif (outSpeaker == '四国めたん') or (outSpeaker == 'めたん'):
            speaker_id = 2 #ノーマル
        elif (outSpeaker == 'ずんだもん'):
            speaker_id = 3 #ノーマル
        elif (outSpeaker == '九州そら') or (outSpeaker == 'そら'):
            speaker_id = 16 #ノーマル
        elif (outSpeaker == 'もち子さん') or (outSpeaker == 'もち子'):
            speaker_id = 20
        elif (outSpeaker == '玄野武宏') or (outSpeaker == '武宏'):
            speaker_id = 11
        elif (outSpeaker == '青山龍星') or (outSpeaker == '龍星'):
            speaker_id = 21
        else:
            return False, None

        # 音声合成
        try:
            host = self.voicevox_host
            port = self.voicevox_port

            params = (
                ('text', text),
                ('speaker', speaker_id),
            )

            res1 = requests.post(
                f'http://{host}:{port}/audio_query',
                params=params,
                timeout=(3, 15)
            )

            if (res1.status_code == 200):
                headers = {'Content-Type': 'application/json',}
                res2 = requests.post(
                    f'http://{host}:{port}/synthesis',
                    headers=headers,
                    params=params,
                    data=json.dumps(res1.json()),
                    timeout=(3, 15)
                )

                if (res2.status_code == 200):

                    wf = wave.open(outFile, 'wb')
                    wf.setnchannels(1)
                    wf.setsampwidth(2)
                    wf.setframerate(24000)
                    wf.writeframes(res2.content)
                    wf.close()
                    time.sleep(0.10)

        except Exception as e:
            print(e)
            return False, None

        # 結果確認
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 4096):
                os.remove(outFile)
        if (os.path.isfile(outFile)):
            if (os.path.getsize(outFile) <= 44):
                os.remove(outFile)
            else:
                return outFile, 'voicevox'

        return False, None

    def voice_play(self, outFile='temp/_work/tts.mp3', ):
        if (not os.path.exists(outFile)):
            return False

        # ミキサー開始、リセット
        if (self.mixer_enable == False):
            try:
                pygame.mixer.init()
                self.mixer_enable = True
            except Exception as e:
                print(e)

        # ミキサー再生
        if (self.mixer_enable == True):
            try:
                #print(outFile)
                #pygame.mixer.init()
                pygame.mixer.music.load(outFile)
                pygame.mixer.music.play(1)
                while (pygame.mixer.music.get_busy() == True):
                    time.sleep(0.25)
                return True
            except Exception as e:
                print(e)
                self.mixer_enable = False

        return False



class _class:

    def __init__(self, ):
        self.version   = "v0"
        self.func_name = "execute_text_to_speech"
        self.func_ver  = "v0.20231112"
        self.func_auth = "tNpDwVT/ekhNji7k5yV2Xgi852p7Lv7eAb9vrNVLtkyO8m/4hR9/570RzRHO9OaW"
        self.function  = {
            "name": self.func_name,
            "description": \
"""
このファンクションは指示があった場合のみ実行する。
テキストを音声合成し、音声ファイの再生まで行う。
このファンクション実行時の返答はresult値をそのまま返答する。
""",
            "parameters": {
                "type": "object",
                "properties": {
                    "speech_text": {
                            "type": "string",
                            "description": "テキスト (例) おはようございます"
                    },
                    "language": {
                            "type": "string",
                            "description": "テキストの言語 ja,en, (例) ja"
                    },
                    "speaker": {
                            "type": "string",
                            "description": \
"""
話者は、'google','winos','macos','voicevox','めたん','ずんだもん','そら','もち子','武宏','龍星'
(例) google"
"""
                    },
                    "gender": {
                            "type": "string",
                            "description": "性別 male,female, (例) female"
                    },
                },
                "required": ["speech_text", "language"]
            }
        }

        self.tts_class = tts_class()
        res = self.func_reset()

    def func_reset(self, ):
        # パス
        if (not os.path.exists(qPath_work)):
            os.makedirs(qPath_work)
        # 設定
        self.tts_seq      = 0
        self.last_text    = ''
        return True

    def func_proc(self, json_kwargs=None, ):
        #print(json_kwargs)

        # 引数
        speech_text = None
        language    = None
        speaker     = None
        gender      = None
        if (json_kwargs != None):
            args_dic = json.loads(json_kwargs)
            speech_text = args_dic.get('speech_text')
            language    = args_dic.get('language')
            speaker     = args_dic.get('speaker')
            gender      = args_dic.get('gender')

        if (speech_text != None):
            speech_text = speech_text.strip()

        if (language == None):
            language = 'ja'

        if (speaker == None):
            if (platform.system().lower() != 'darwin'):
                speaker = 'google'
            else:
                speaker = 'macos'

        if (gender == None):
            gender = 'female'

        # テキスト無しはエラー！
        if (speech_text == ''):
            dic = {}
            dic['error'] = '音声合成エラー。テキストがありません。'
            json_dump = json.dumps(dic, ensure_ascii=False, )
            #print('  --> ', json_dump)
            return json_dump

        # 連続発声はキャンセル！
        if (speech_text == self.last_text):
            dic = {}
            dic['result'] = 'さきほど音声合成及び音声ファイルの再生もしました'
            json_dump = json.dumps(dic, ensure_ascii=False, )
            #print('  --> ', json_dump)
            return json_dump

        # 出力ファイル
        self.tts_seq += 1
        if (self.tts_seq > 9999):
            self.tts_seq = 1
        seq4 = '{:04}'.format(self.tts_seq)
        outFile = qPath_work + self.func_name + '.' + seq4 + '.mp3'

        if (os.path.isfile(outFile)):
            try:
                os.remove(outFile)
            except Exception as e:
                print(e)
        if (os.path.isfile(outFile[:-4] + '.wav')):
            try:
                os.remove(outFile[:-4] + '.wav')
            except Exception as e:
                print(e)

        # 初期化
        tts_ok  = False
        play_ok = False

        # 音声合成 (voicevox)
        if (tts_ok == False):
            if (speaker.lower() == 'voicevox') \
            or (speaker == '四国めたん') or (speaker == 'めたん') \
            or (speaker == 'ずんだもん') \
            or (speaker == '九州そら') or (speaker == 'そら') \
            or (speaker == 'もち子さん') or (speaker == 'もち子') \
            or (speaker == '玄野武宏') or (speaker == '武宏') \
            or (speaker == '青山龍星') or (speaker == '龍星'):
                res, api = self.tts_class.voicevox_tts(outText=speech_text, outLang=language, outSpeaker=speaker, outGender=gender, outFile=outFile, )
                if (res != False):
                    tts_ok  = True
                    outFile = res

        # 音声合成 (winos)
        if (tts_ok == False):
            if (speaker.lower() == 'winos'):
                res, api = self.tts_class.winos_tts(outText=speech_text, outLang=language, outSpeaker=speaker, outGender=gender, outFile=outFile, )
                if (res != False):
                    tts_ok  = True
                    outFile = res

        # 音声合成 (macos)
        if (tts_ok == False):
            if (speaker.lower() == 'macos') \
            or (speaker.lower() == 'kyoko') \
            or (speaker.lower() == 'otoya'):
                res, api = self.tts_class.macos_tts(outText=speech_text, outLang=language, outSpeaker=speaker, outGender=gender, outFile=outFile, )
                if (res != False):
                    tts_ok  = True
                    outFile = res

        # 音声合成 (google)
        if (tts_ok == False):
            #if (speaker.lower() == 'google'):
                res, api = self.tts_class.google_tts(outText=speech_text, outLang=language, outSpeaker=speaker, outGender=gender, outFile=outFile, )
                if (res != False):
                    if (speaker.lower() != 'google'):
                        print('tts : ' + speaker + ' -> google ok')
                    tts_ok  = True
                    outFile = res
                else:
                    if (speaker.lower() != 'google'):
                        print('tts : ' + speaker + ' -> google ng!')

        # 音声合成エラー
        if (tts_ok == False):
            dic = {}
            dic['error'] = '音声合成でエラーが発生しました'
            json_dump = json.dumps(dic, ensure_ascii=False, )
            #print('  --> ', json_dump)
            return json_dump

        # 音声再生
        if (tts_ok == True):
            print('tts  :', outFile, '(' + api + ')')

            self.last_text = speech_text

            print('play :', outFile)
            res = self.tts_class.voice_play(outFile=outFile, )
            if (res != False):
                play_ok = True

        # 戻り値
        if (play_ok == True):
            dic = {}
            dic['result'] = '音声合成及び音声ファイルの再生もしました'
            dic['output_path'] = outFile
            json_dump = json.dumps(dic, ensure_ascii=False, )
            #print('  --> ', json_dump)
            return json_dump
        else:
            dic = {}
            dic['error'] = '音声ファイルの再生でエラーが発生しました'
            json_dump = json.dumps(dic, ensure_ascii=False, )
            #print('  --> ', json_dump)
            return json_dump

if __name__ == '__main__':

    ext = _class()
    print(ext.func_proc( '{ "speech_text": "おはようございます", "language": "ja", "speaker": "voicevox", "gender": "female" }' ))
    print(ext.func_proc( '{ "speech_text": "よろしくお願いします", "language": "ja", "speaker": "winos", "gender": "female" }' ))
    print(ext.func_proc( '{ "speech_text": "ありがとうございます", "language": "ja", "speaker": "macos", "gender": "female" }' ))
    print(ext.func_proc( '{ "speech_text": "お疲れさまでした", "language": "ja", "speaker": "google", "gender": "female" }' ))
